from unittest import TestCase
import mock
from jav.core.javTime import Time
from datetime import datetime
import pytz
from cement.core import foundation

class TestTime(TestCase):

    @classmethod
    def get_app(self):
        """App init, necessary to get to the logging service"""
        app = foundation.CementApp('myapp')
        app.setup()
        app.run()
        return app

    def test_get_current_date(self):
        answer_date = datetime.now(pytz.timezone('America/Toronto')).strftime('%Y-%m-%d')
        response_date = Time.get_current_date().strftime('%Y-%m-%d')
        self.assertEqual(answer_date, response_date)

    @mock.patch('jav.core.javConfig')
    def test_get_end_date(self, mock_config):
        mock_config.get_config_value = mock.MagicMock(return_value='2017-01-22')

        # App init, necessary to get to the logging service
        app = self.get_app()

        time = Time(app.log, mock_config)

        end_date_answer = datetime.strptime('2017-01-22', "%Y-%m-%d").date()
        end_date_response = time.get_end_date()

        self.assertEqual(end_date_answer, end_date_response)






import requests

class Jira(object):
    """
        Handle connection to JIRA
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config

    def test_jira_result(self, jira_query, jira_result):
        if 'issues' not in jira_result:
            self.log.error('JIRA did not return an array of issues, your JQL query appears to be incorrect')
            self.log.error(jira_query)
            self.log.error(jira_result['errorMessages'])
            exit()
        else:
            return True

    def get_completed_tickets(self, date_current):
        self.log.info(
            'Jira.get_completed_tickets(): Getting completed tickets from Jira for date: ' + date_current.strftime(
                "%Y-%m-%d"))
        jira_query = self.config.get_config_value('jira_jql_velocity') + ' ON(\"' + date_current.strftime(
            "%Y-%m-%d") + '\")'
        result = self.call(jira_query).json()
        if self.test_jira_result(jira_query, result):
            return result

    def get_remaining_tickets(self):
        self.log.info('Jira.get_remaining_tickets(): Getting remaining work tickets from Jira')
        result = self.call(self.config.get_config_value('jira_jql_remaining')).json()
        if self.test_jira_result(self.config.get_config_value('jira_jql_remaining'), result):
            return result

    def call(self, jql_query):
        headers = {
            "Content-Type": "application/json",
        }

        fields = [
            'assignee'
            , 'issuetype'
        ]
        if self.config.get_config_value('jira_field_points') is not None:
            fields.append(self.config.get_config_value('jira_field_points'))

        params = (
            ("jql", jql_query),
            ("startAt", 0),
            ("maxResults", 1500),
            ("fields", fields),
        )
        try:
            tickets = requests.get(self.config.get_config_value('jira_host') + "/rest/api/2/search", headers=headers,
                                   params=params,
                                   auth=(
                                       self.config.get_config_value('jira_username'),
                                       self.config.get_config_value('jira_password')))
            return tickets
        except Exception as ex:
            # ConnectionError
            template = "An exception of type {0} occurred. Arguments:\n{1!r}"
            message = template.format(type(ex).__name__, ex.args)
            self.log.error('Jira.call(): Unable to communicate with remote JIRA server, exiting... ')
            self.log.error(message)
            exit()



"""Jira Agile Velocity bootstrapping."""

# All built-in application controllers should be imported, and registered
# in this file in the same way as javBaseController.

from jav.cli.controllers.base import javBaseController

def load(app):
    app.handler.register(javBaseController)



import numpy as np
import pandas as pd

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

from bokeh.charts import output_file, save, show
from bokeh.layouts import layout
from bokeh.plotting import figure, ColumnDataSource
from bokeh.models import HoverTool, OpenURL, TapTool, FactorRange
from jav.core.javTime import Time


class BuildChart(object):
    def __init__(self, log, config):
        self.log = log
        self.config = config
        self.time = Time(self.log, self.config)

    def build_url_completed(self, current_date):
        return \
            self.config.get_config_value('jira_host') + '/issues/?jql=' \
            + quote(
                self.config.get_config_value('jira_jql_velocity')
                + ' ON(\"'
                + current_date.strftime("%Y-%m-%d")
                + '\")'
            )

    def build_url_remaining(self, jira_field, current_type):
        return \
            self.config.get_config_value('jira_host') + '/issues/?jql=' \
            + quote(
                self.config.get_config_value('jira_jql_remaining')
                + ' AND '
                + jira_field
                + ' = \"'
                + current_type
                + '\"'
            )

    def build_velocity_days(self, stats_data):
        self.log.info('Generating graph about daily effort')
        # Velocity Days
        x_dates = []
        y_values = []
        y_avg = []
        x_dates_display = []
        jira_url = []
        for scan_day in stats_data:
            x_dates.append(stats_data[scan_day]['datetime'])
            x_dates_display.append(stats_data[scan_day]['datetime'].strftime('%a %b %d, %Y'))
            y_values.append(stats_data[scan_day][self.config.get_config_value('stats_metric')])
            jira_url.append(self.build_url_completed(stats_data[scan_day]['datetime']))
            week_idx = '4'
            if week_idx not in stats_data[scan_day]['anyday']:
                week_idx = 4
            if 'avg' in stats_data[scan_day]['anyday'][week_idx]:
                y_avg.append(stats_data[scan_day]['anyday'][week_idx]['avg'])
            else:
                y_avg.append(0)
        # prepare some data
        data_dates = np.array(x_dates, dtype=np.datetime64)
        source = ColumnDataSource(
            data=dict(
                x_data_dates=data_dates,
                y_values=y_values,
                y_avg_values=y_avg,
                x_dates_display=x_dates_display,
                jira_url=jira_url
            )
        )
        # Declare tools
        hover = HoverTool(
            tooltips=[
                ('Date', '@x_dates_display'),
                (self.config.get_config_value('stats_metric').capitalize(), '@y_values'),
                ('Average', '@y_avg_values'),
            ]
        )
        # create a new plot with a a datetime axis type
        p = figure(width=1000, height=350, x_axis_type='datetime',
                   tools=['save', 'pan', 'box_zoom', 'reset', hover, 'tap'])
        if self.config.get_config_value('stats_metric') == 'tickets':
            metric_legend = 'Tickets'
        else:
            metric_legend = 'Story Points'
        # add renderers
        p.circle('x_data_dates', 'y_values', size=4, color='red', alpha=0.4, legend=metric_legend, source=source)
        p.line('x_data_dates', 'y_avg_values', color='blue', legend='4 Weeks Average', source=source)
        taptool = p.select(type=TapTool)
        taptool.callback = OpenURL(url='@jira_url')
        # NEW: customize by setting attributes
        p.title.text = 'Daily Velocity'
        p.legend.location = 'top_left'
        p.grid.grid_line_alpha = 0
        p.xaxis.axis_label = 'Days (Monday-Friday)'
        p.yaxis.axis_label = metric_legend
        p.ygrid.band_fill_color = 'olive'
        p.ygrid.band_fill_alpha = 0.1

        return p

    def build_velocity_weeks(self, stats_data):
        self.log.info('Generating graph about weekly effort')
        # Velocity Days
        x_dates = []
        y_values = []
        y_avg = []
        y_weeks = []
        daily_avg = []
        for scan_day in stats_data:
            x_dates.append(stats_data[scan_day]['datetime'])
            y_values.append(stats_data[scan_day][self.config.get_config_value('stats_metric')])
            y_weeks.append(stats_data[scan_day]['weektxt'])
            daily_avg.append(round(float(stats_data[scan_day][self.config.get_config_value('stats_metric')]) / 5, 1))
            if len(stats_data[scan_day]['stats']) > 0:
                week_idx = '4'
                if week_idx not in stats_data[scan_day]['stats']:
                    week_idx = 4
                if 'avg' in stats_data[scan_day]['stats'][week_idx]:
                    y_avg.append(stats_data[scan_day]['stats'][week_idx]['avg'])
                else:
                    y_avg.append(0)
        # prepare some data
        data_dates = np.array(x_dates, dtype=np.datetime64)
        source = ColumnDataSource(
            data=dict(
                x_data_dates=data_dates,
                y_values=y_values,
                y_avg_values=y_avg,
                y_weeks=y_weeks,
                daily_avg=daily_avg,
            )
        )
        # Declare tools
        hover = HoverTool(
            tooltips=[
                ('Week', '@y_weeks'),
                (self.config.get_config_value('stats_metric').capitalize(), '@y_values'),
                ('Avg ' + self.config.get_config_value('stats_metric') + ' per week', '@y_avg_values'),
                ('Avg ' + self.config.get_config_value('stats_metric') + ' per day', '@daily_avg'),
            ]
        )
        # create a new plot with a a datetime axis type
        p = figure(width=1000, height=350, x_axis_type='datetime', tools=['save', 'pan', 'box_zoom', 'reset', hover])
        if self.config.get_config_value('stats_metric') == 'tickets':
            metric_legend = 'Tickets'
        else:
            metric_legend = 'Story Points'
        # add renderers
        p.circle('x_data_dates', 'y_values', size=4, color='red', alpha=0.4, legend=metric_legend, source=source)
        p.line('x_data_dates', 'y_avg_values', color='blue', legend='4 Weeks Average', source=source)
        # NEW: customize by setting attributes
        p.title.text = 'Weekly Velocity'
        p.legend.location = 'top_left'
        p.grid.grid_line_alpha = 0
        p.xaxis.axis_label = 'Weeks'
        p.yaxis.axis_label = metric_legend
        p.ygrid.band_fill_color = 'olive'
        p.ygrid.band_fill_alpha = 0.1

        return p

    def get_points_per_type(self, stats_data):
        plot_values = []
        for scan_day in stats_data:
            total_metric = stats_data[scan_day][self.config.get_config_value('stats_metric')]
            for assignee in stats_data[scan_day]['types']:
                plot_values.append([
                    stats_data[scan_day]['types'][assignee]['type']
                    , stats_data[scan_day]['types'][assignee][self.config.get_config_value('stats_metric')]
                    , self.build_url_remaining('type', stats_data[scan_day]['types'][assignee]['type'])
                ])
            break
        return plot_values, total_metric

    @staticmethod
    def get_remaining_source(dat):
        source = ColumnDataSource(dict(
            x_values=dat.entity
            , y_values=dat.value
            , jira_url=dat.jira_url
        ))
        return source

    def chart_remaining_types(self, stats_data):
        self.log.info('Generating graph about remaining effort per ticket type')
        plot_values, total_metric = self.get_points_per_type(stats_data)
        dat = pd.DataFrame(plot_values, columns=['entity', 'value', 'jira_url'])
        source = self.get_remaining_source(dat)
        # Declare tools
        hover = HoverTool(
            tooltips=[
                ('Type:', '@x_values'),
                ('Points:', '@y_values')
            ]
        )
        plot = figure(
            plot_width=600
            , plot_height=300
            , x_axis_label='Ticket Type'
            , y_axis_label=self.config.get_config_value('stats_metric')
            , title='Remaining ' + self.config.get_config_value('stats_metric') + ' per Ticket Type (Total: ' + str(
                total_metric) + ')'
            , x_range=FactorRange(factors=list(dat.entity))
            , tools=['save', 'pan', 'box_zoom', 'reset', hover, 'tap']
        )
        taptool = plot.select(type=TapTool)
        taptool.callback = OpenURL(url='@jira_url')
        plot.vbar(source=source, x='x_values', top='y_values', bottom=0, width=0.3, color='green')
        return plot

    def get_points_per_assignee(self, stats_data):
        plot_values = []
        for scan_day in stats_data:
            total_metric = stats_data[scan_day][self.config.get_config_value('stats_metric')]
            for assignee in stats_data[scan_day]['assignees']:
                plot_values.append([
                    stats_data[scan_day]['assignees'][assignee]['displayName']
                    , stats_data[scan_day]['assignees'][assignee][self.config.get_config_value('stats_metric')]
                    , self.build_url_remaining('assignee', stats_data[scan_day]['assignees'][assignee]['displayName'])
                ])
            break
        return plot_values, total_metric

    def chart_remaining_assignees(self, stats_data):
        self.log.info('Generating graph about remaining effort per assignee')
        plot_values, total_metric = self.get_points_per_assignee(stats_data)

        dat = pd.DataFrame(plot_values, columns=['entity', 'value', 'jira_url'])
        source = self.get_remaining_source(dat)
        # Declare tools
        hover = HoverTool(
            tooltips=[
                ('Assignee:', '@x_values'),
                ('Points:', '@y_values')
            ]
        )
        plot = figure(
            plot_width=600
            , plot_height=300
            , x_axis_label='Assignee'
            , y_axis_label=self.config.get_config_value('stats_metric')
            , title='Remaining ' + self.config.get_config_value('stats_metric') + ' per Jira Assignee (Total: ' + str(
                total_metric) + ')'
            , x_range=FactorRange(factors=list(dat.entity))
            , tools=['save', 'pan', 'box_zoom', 'reset', hover, 'tap']
        )
        taptool = plot.select(type=TapTool)
        taptool.callback = OpenURL(url='@jira_url')

        plot.vbar(source=source, x='x_values', top='y_values', bottom=0, width=0.3, color='blue')
        return plot

    def chart_remaining_days(self, stats_data):
        self.log.info('Generating graph about estimated remaining days')
        plot_values = []
        for scan_day in stats_data:
            for rol_avg in stats_data[scan_day]['days_to_completion']:
                if rol_avg == 'current':
                    col_txt = 'This Week'
                elif rol_avg == 'all':
                    col_txt = 'All Time'
                else:
                    col_txt = str(rol_avg) + ' Weeks'
                plot_values.append([
                    col_txt
                    , stats_data[scan_day]['days_to_completion'][rol_avg]
                ])
            break
        dat = pd.DataFrame(plot_values, columns=['entity', 'value'])
        source = ColumnDataSource(dict(
            x_values=dat.entity
            , y_values=dat.value
        ))
        # Declare tools
        hover = HoverTool(
            tooltips=[
                ('Estimated based on:', '@x_values'),
                ('Days to completion:', '@y_values')
            ]
        )
        plot = figure(
            plot_width=600
            , plot_height=300
            , x_axis_label='Estimations calculated based on'
            , y_axis_label='Remaining days'
            , title='Estimated days to completion'
            , x_range=FactorRange(factors=list(dat.entity))
            , tools=['save', 'pan', 'box_zoom', 'reset', hover]
        )
        plot.vbar(source=source, x='x_values', top='y_values', bottom=0, width=0.3, color='red')
        return plot

    def main(self, stats_days, stats_weeks, stats_remaining):
        days_chart = self.build_velocity_days(stats_days)
        weeks_chart = self.build_velocity_weeks(stats_weeks)

        remaining_types = self.chart_remaining_types(stats_remaining)
        remaining_assignees = self.chart_remaining_assignees(stats_remaining)
        remaining_days = self.chart_remaining_days(stats_remaining)

        bokeh_layout = layout([
            [days_chart]
            , [weeks_chart]
            , [remaining_types, remaining_assignees, remaining_days]
        ], sizing_mode='stretch_both')

        # output to static HTML file
        output_file(self.config.filepath_charts + 'index.html',
                    title='[' + self.time.get_current_date().strftime(
                        "%Y-%m-%d") + '] - Jira Metrics, built on: ' + self.time.get_current_date().strftime(
                        "%Y-%m-%d"))
        save(bokeh_layout)
        show(bokeh_layout)



import os


class Clear(object):
    """
        Classe used to clear previously processed data
    """

    def __init__(self, log, config):
        self.log = log
        self.__config = config

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, config_class):
        self.__config = config_class

    def main(self):
        self.log.info('Clearing previously downloaded/processed data')
        self.log.info(self.config.filepath_data_completion)
        if os.path.isfile(self.config.filepath_data_completion):
            os.remove(self.config.filepath_data_completion)
            self.log.info(self.config.filepath_data_completion + ' Removed')

        return True



"""Jira Agile Velocity exception classes."""

class javError(Exception):
    """Generic errors."""
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg

class javConfigError(javError):
    """Config related errors."""
    pass

class javRuntimeError(javError):
    """Generic runtime errors."""
    pass

class javArgumentError(javError):
    """Argument related errors."""
    pass



"""Testing utilities for Jira Agile Velocity."""

from jav.cli.main import javTestApp
from cement.utils.test import *

class javTestCase(CementTestCase):
    app_class = javTestApp

    def setUp(self):
        """Override setup actions (for every test)."""
        super(javTestCase, self).setUp()

    def tearDown(self):
        """Override teardown actions (for every test)."""
        super(javTestCase, self).tearDown()




class LogConfig(object):
    """
        Classe used to clear previously processed data
    """

    def __init__(self, log, app_config, log_filepath):
        self.log = log

        # This section is used to set-up log files
        self.app_config = app_config
        self.app_config.set(self.log._meta.config_section, 'file', log_filepath)
        self.app_config.set(self.log._meta.config_section, 'rotate', True)
        self.app_config.set(self.log._meta.config_section, 'max_bytes', 512000)
        self.app_config.set(self.log._meta.config_section, 'max_files', 10)
        self.log._setup_file_log()






import yaml
from os.path import expanduser
import os
import json
import jsonschema
import copy


class Config(object):
    """
        Class in charge to configuration management
    """

    def __init__(self, log, config_path = expanduser('~') + '/.jav/', config_values = None):
        self.log = log
        self.__config_path = self.prep_config_path(config_path)
        self.__config_filename = 'config.yml'
        self.__config_filepath = self.__config_path + self.__config_filename

        self.__filepath_data_completion = self.__config_path + 'data_completion.jsonl'
        self.__filepath_data_remaining = self.__config_path + 'data_remaining.json'

        self.__filepath_stats_days = self.__config_path + 'stats_days.jsonl'
        self.__filepath_stats_weeks = self.__config_path + 'stats_weeks.jsonl'
        self.__filepath_stats_remaining = self.__config_path + 'stats_remaining.jsonl'

        self.__filepath_charts = self.prep_config_path(self.__config_path + 'built-charts/')

        # config_init is used to record if the init method was once called
        self.__config_init = False

        self.__config = config_values
        self.__schema = {
            '$schema': 'http://json-schema.org/draft-04/schema#'
            , 'title': 'configObj'
            , 'description': 'Configuration values '
            , 'type': 'object'
            , 'additionalProperties': False
            , 'properties': {
                'end_date': {
                    'type': ['string']
                    , 'description': 'Earliest date to collect data from'
                    , 'default': '2017-01-01'
                }
                , 'jira_username': {
                    'type': ['string']
                    , 'description': 'JIRA Username used for the API call'
                    , 'default': 'USERNAME'
                }
                , 'jira_password': {
                    'type': ['string']
                    , 'description': 'JIRA Password used for the API call'
                    , 'default': 'PASSWORD'
                }
                , 'jira_host': {
                    'type': ['string']
                    , 'description': 'Remote JIRA Hostname (https://jira.yourdomain.tld)'
                    , 'default': 'HOST'
                }
                , 'jira_jql_velocity': {
                    'type': ['string']
                    , 'description': 'JIRA JQL Query used to track story points completion '
                                     '(for example: project = WORK and status changed to Done)'
                    , 'default': 'JQL QUERY VELOCITY'
                }
                , 'jira_jql_remaining': {
                    'type': ['string']
                    , 'description': 'JIRA JQL Query used to determine remaining work '
                                     '(for example: sprint in openSprints() and type in (Story, Task, Defect))'
                    , 'default': 'JQL QUERY REMAINING'
                }
                , 'jira_field_points': {
                    'type': ['string']
                    , 'description': 'Field used by JIRA to store Story Points'
                    , 'default': 'customfield_10002'
                }
                , 'slack_channel': {
                    'type': ['string', 'null']
                    , 'description': 'Slack channel to post this message to'
                    , 'default': '#general'
                }
                , 'slack_webhook': {
                    'type': ['string', 'null']
                    , 'description': 'Slack webhook (https://api.slack.com/incoming-webhooks)'
                    , 'default': 'SLACK WEBHOOK'
                }
                , 'rolling_stats': {
                    'type': 'array'
                    , 'description': 'Array of weeks to group stats by. For example average over the past 4 weeks, past 8 weeks etc...'
                    , 'default': [4, 8]
                }
                , 'stats_metric': {
                    'type': 'string'
                    , 'description': 'Type of metric to be used to calculate stats, build charts and messages (points or tickets)'
                    , 'default': 'points'
                }
                , 'git_repo': {
                    'type': ['string', 'null']
                    , 'description': 'Git repository to be used to publish charts'
                    , 'default': 'https://github.com/Fgerthoffert/test-agile-page.git'
                }
                , 'git_branch': {
                    'type': ['string', 'null']
                    , 'description': 'Git branch for the charts'
                    , 'default': 'gh-pages'
                }
                , 'git_localpath': {
                    'type': ['string', 'null']
                    , 'description': 'Path of the git directory on the local filesystem'
                    , 'default': '/Users/francois/Desktop/test-agile-page/'
                }
                , 'git_pathdirectory': {
                    'type': ['string', 'null']
                    , 'description': 'Directory path, within localpath to be used to copy files'
                    , 'default': 'charts/'
                }
                , 'git_pageurl': {
                    'type': ['string', 'null']
                    , 'description': 'URL of the page to be sent via slack'
                    , 'default': 'https://fgerthoffert.github.io/test-agile-page/charts/'
                }
            }
        }

        if self.config is not None:
            self.log.info('Default configuration provided as part of class initialization')
        elif os.path.isdir(self.__config_path) is False or os.path.isfile(self.__config_filepath) is False:
            self.__config_init = self.init_config()
        elif os.path.isfile(self.__config_filepath):
            self.load_config()

    @staticmethod
    def prep_config_path(config_path):
        if config_path is None:
            config_path = expanduser('~') + '/.jav/'
        elif config_path[-1] != '/':
            config_path = config_path + '/'
        if not os.path.isdir(config_path):
            try:
                os.makedirs(config_path)
            except Exception as ex:
                print('Directory: ' + config_path + ' already exists')
                print(ex.message)
        return config_path

    @property
    def schema(self):
        return self.__schema

    @property
    def config_init(self):
        return self.__config_init

    @property
    def filepath_data_completion(self):
        return self.__filepath_data_completion

    @property
    def filepath_data_remaining(self):
        return self.__filepath_data_remaining

    @property
    def filepath_stats_days(self):
        return self.__filepath_stats_days

    @property
    def filepath_stats_weeks(self):
        return self.__filepath_stats_weeks

    @property
    def filepath_stats_remaining(self):
        return self.__filepath_stats_remaining

    @property
    def filepath_charts(self):
        return self.__filepath_charts

    @property
    def config(self):
        return self.__config

    @config.setter
    def config(self, config_obj):
        try:
            jsonschema.validate(config_obj, self.schema)
        except Exception as ex:
            # ValidationError
            self.log.error('The requested configuration update is not valid according to the schema')
            template = 'An exception of type {0} occurred. Arguments:\n{1!r}'
            message = template.format(type(ex).__name__, ex.args)
            self.log.error(message)
            exit()

        self.__config = config_obj

    @property
    def config_path(self):
        return self.__config_path

    def get_config_value(self, key):
        """Get a single value of the dictionary"""
        return self.config[key]

    def set_config_value(self, key, value):
        """Updates a key/value pair in the config, verify it against the schema before updating config"""
        value_init = None
        if self.config is not None and key in self.config:
            value_init = self.config[key]
        config_tmp = copy.deepcopy(self.config)
        config_tmp[key] = value
        self.config = config_tmp
        self.log.info(
            'Updated config: ' + key + ' from: ' + str(value_init) + ' to: ' + str(value))
        return self.config[key]

    def load_config(self):
        """Load content of the YML config file"""
        self.log.info('Config.open_config: Opening settings from config file: ' + self.__config_filepath)
        with open(self.__config_filepath, 'r') as ymlfile:
            cfg = yaml.safe_load(ymlfile)
        self.config = cfg
        return self.config

    def write_config(self):
        """Write config into a YML config file"""
        self.log.info('Config.write_config(): Writing config file to: ' + self.__config_filepath)
        self.log.info('Config.write_config(): Content:' + json.dumps(self.config))
        if not os.path.isdir(self.__config_path):
            os.mkdir(self.__config_path)

        config_file = open(self.__config_filepath, 'w')
        yaml.dump(self.__config, config_file, default_flow_style=False)

        return self.config

    def init_config(self):
        """Manually initialize configuration"""
        self.log.info('Unable to find config file, initializing')
        self.log.info('Press Enter for [default], CTRL+C to exit')
        config_schema = self.schema['properties']
        for config_key in sorted(config_schema):
            config_value = self.init_config_value(config_key)
            self.set_config_value(config_key, config_value)
        self.write_config()
        return True

    def init_config_auto(self):
        """Automatically initializing configuration, setting all values to default"""
        config_schema = self.schema['properties']
        for config_key in sorted(config_schema):
            self.set_config_value(config_key, config_schema[config_key]['default'])
        return True

    def init_config_value(self, config_key):
        value_default = None
        value_current = None
        value_suggested = None
        if 'default' in self.schema['properties'][config_key]:
            value_default = self.schema['properties'][config_key]['default']

        if self.config is not None and config_key in self.config:
            value_current = self.config[config_key]

        self.log.info(config_key + ' - ' + self.schema['properties'][config_key]['description'])
        self.log.info('Current Value: ' + str(value_current) + ' Default Value: ' + str(value_default))
        if value_current is None and value_default is not None:
            value_suggested = value_default
        elif value_current is not None:
            value_suggested = value_current

        # Note: https://stackoverflow.com/questions/10885537/raw-input-has-been-eliminated-from-python-3-2
        # https://stackoverflow.com/questions/21731043/use-of-input-raw-input-in-python-2-and-3
        try:
            input = raw_input
        except NameError:
            pass

        try:
            config_value = input('[' + str(value_suggested) + ']:')
        except Exception as ex:
            # If exception, we consider the value to be empty (fallback to default)
            self.log.debug(ex.message)
            config_value = ''

        self.log.debug('Received value: ' + config_value)

        if config_value == '' and value_suggested is not None:
            config_value = value_suggested
        elif config_value == '' and 'null' not in self.schema['properties'][config_key]['type']:
            self.log.warn('This value cannot be empty, please enter a value')
            config_value = self.init_config_value(config_key)

        if 'string' in self.schema['properties'][config_key]['type']:
            config_value = str(config_value)
        elif 'number' in self.schema['properties'][config_key]['type']:
            config_value = int(config_value)

        return config_value




from unittest import TestCase
import mock
from jav.core.javImportData import ImportData
from jav.core.javConfig import Config
from jav.core.javJira import Jira
from jav.core.javFiles import Files
from datetime import datetime
import collections
from cement.core import foundation

class TestImportData(TestCase):

    @classmethod
    def get_app(self):
        """App init, necessary to get to the logging service"""
        app = foundation.CementApp('myapp')
        app.setup()
        app.run()
        return app

    @classmethod
    def get_jira_issues(self):
        """Return a simple list of issues simulating a JIRA response"""
        jira_issues = {
            'issues': [
                {'fields': {'issuetype': {'name': 'defect'}, 'jira_points_field': 10, 'assignee': {'name': 'johnd', 'displayName': 'John Doe'}}}
                , {'fields': {'issuetype': {'name': 'defect'}, 'jira_points_field': 20, 'assignee': {'name': 'brucew', 'displayName': 'Bruce Wayne'}}}
                , {'fields': {'issuetype': {'name': 'defect'}, 'assignee': {'name': 'darthv', 'displayName': 'Darth Vader'}}}
                , {'fields': {'issuetype': {'name': 'story'}, 'jira_points_field': 2, 'assignee': {'name': 'brucew', 'displayName': 'Bruce Wayne'}}}
                , {'fields': {'issuetype': {'name': 'story'}, 'jira_points_field': 5, 'assignee': {'name': 'darthv', 'displayName': 'Darth Vader'}}}
            ]
        }
        return jira_issues

    @classmethod
    def get_data_completion(self):
        data = collections.OrderedDict()
        data['20170525'] = {
            'datetime': datetime.strptime('2017-05-25', '%Y-%m-%d').date()
            , 'tickets': 10
            , 'points': 15
            , 'assignees': {
                'johnd': {'tickets': 2, 'points': 3, 'displayName': 'John Doe'}
                , 'brucew': {'tickets': 3, 'points': 7, 'displayName': 'Bruce Wayne'}
                , 'darthv': {'tickets': 5, 'points': 5, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'Defect': {'tickets': 7, 'points': 1, 'type': 'Defect'}
                , 'Story': {'tickets': 1, 'points': 10, 'type': 'Story'}
                , 'Task': {'tickets': 2, 'points': 4, 'type': 'Task'}
            }
        }
        data['20170524'] = {
            'datetime': datetime.strptime('2017-05-24', '%Y-%m-%d').date()
            , 'tickets': 4
            , 'points': 30
            , 'assignees': {
                'johnd': {'tickets': 1, 'points': 10, 'displayName': 'John Doe'}
                , 'darthv': {'tickets': 3, 'points': 20, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'Defect': {'tickets': 2, 'points': 14, 'type': 'Defect'}
                , 'Story': {'tickets': 2, 'points': 16, 'type': 'Story'}
            }
        }
        data['20170523'] = {
            'datetime': datetime.strptime('2017-05-23', '%Y-%m-%d').date()
            , 'tickets': 8
            , 'points': 20
            , 'assignees': {
                'johnd': {'tickets': 2, 'points': 5, 'displayName': 'John Doe'}
                , 'brucew': {'tickets': 2, 'points': 10, 'displayName': 'Bruce Wayne'}
                , 'darthv': {'tickets': 4, 'points': 5, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'Defect': {'tickets': 1, 'points': 7, 'type': 'Defect'}
                , 'Story': {'tickets': 3, 'points': 19, 'type': 'Story'}
                , 'Task': {'tickets': 4, 'points': 4, 'type': 'Task'}
            }
        }
        return data

    @classmethod
    def get_data_completion_answer(self):
        data = collections.OrderedDict()
        data['20170526'] = {
            'datetime': datetime.strptime('2017-05-26', '%Y-%m-%d').date()
            , 'tickets': 5
            , 'points': 37
            , 'assignees': {
                'johnd': {'tickets': 1, 'points': 10, 'displayName': 'John Doe'}
                , 'brucew': {'tickets': 2, 'points': 22, 'displayName': 'Bruce Wayne'}
                , 'darthv': {'tickets': 2, 'points': 5, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'defect': {'tickets': 3, 'points': 30, 'type': 'defect'}
                , 'story': {'tickets': 2, 'points': 7, 'type': 'story'}
            }
        }
        data['20170525'] = self.get_data_completion()['20170525']
        data['20170524'] = self.get_data_completion()['20170524']
        data['20170523'] = self.get_data_completion()['20170523']
        data['20170522'] = {
            'datetime': datetime.strptime('2017-05-22', '%Y-%m-%d').date()
            , 'tickets': 5
            , 'points': 37
            , 'assignees': {
                'johnd': {'tickets': 1, 'points': 10, 'displayName': 'John Doe'}
                , 'brucew': {'tickets': 2, 'points': 22, 'displayName': 'Bruce Wayne'}
                , 'darthv': {'tickets': 2, 'points': 5, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'defect': {'tickets': 3, 'points': 30, 'type': 'defect'}
                , 'story': {'tickets': 2, 'points': 7, 'type': 'story'}
            }
        }
        return data

    @classmethod
    def get_remaining_work_answer(self):
        response = {
            'tickets': 5
            , 'points': 37
            , 'assignees': {
               'johnd': {'tickets': 1, 'points': 10, 'displayName': 'John Doe'}
               , 'brucew': {'tickets': 2, 'points': 22, 'displayName': 'Bruce Wayne'}
               , 'darthv': {'tickets': 2, 'points': 5, 'displayName': 'Darth Vader'}
            }
            , 'types': {
                'defect': {'tickets': 3, 'points': 30, 'type': 'defect'}
                , 'story': {'tickets': 2, 'points': 7, 'type': 'story'}
            }
        }
        return response

    @mock.patch('jav.core.javConfig')
    def test_get_story_point(self, mock_config):
        #mock_config.get_config_value.return_value = 'jira_points_field'
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')

        # App init, necessary to get to the logging service
        app = self.get_app()

        # Init the Import Data Class
        import_data = ImportData(app.log, mock_config)

        # If field exists and send 1, should return 1
        self.assertEqual(import_data.get_story_points({'fields':{'jira_points_field': 1}}), 1)
        # If field does not exist and send 1, should return none
        self.assertEqual(import_data.get_story_points({'fields':{'does_not_exist': 1}}), None)


    @mock.patch('jav.core.javConfig')
    def test_story_types_count(self, mock_config):
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')

        # App init, necessary to get to the logging service
        app = self.get_app()

        # Init the Import Data Class
        import_data = ImportData(app.log, mock_config)

        # Send a couple of issues and ensure returned value are correct
        self.assertEqual(import_data.story_types_count(self.get_jira_issues()['issues']), {'story': {'tickets': 2, 'points': 7, 'type': 'story'}, 'defect': {'tickets': 3, 'points': 30, 'type': 'defect'}})

    @mock.patch('jav.core.javConfig')
    def test_count_story_points(self, mock_config):
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')
        points_per_story = 3
        expected_result = points_per_story*len(self.get_jira_issues()['issues'])

        # App init, necessary to get to the logging service
        app = self.get_app()

        # Init the Import Data Class
        import_data = ImportData(app.log, mock_config)
        import_data.get_story_points = mock.MagicMock(return_value=points_per_story)
        self.assertEqual(import_data.count_story_points(self.get_jira_issues()['issues']), expected_result)

    @mock.patch('jav.core.javConfig')
    def test_assignee_count(self, mock_config):
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')
        # App init, necessary to get to the logging service
        app = self.get_app()

        # Init the Import Data Class
        import_data = ImportData(app.log, mock_config)

        assignee_count_answer = {
            'johnd': {'displayName': 'John Doe', 'points': 10, 'tickets': 1}
            , 'brucew': {'displayName': 'Bruce Wayne', 'points': 22, 'tickets': 2}
            , 'darthv': {'displayName': 'Darth Vader', 'points': 5, 'tickets': 2}
        }
        assignee_count_response = import_data.assignee_count(self.get_jira_issues()['issues'])

        # Send a couple of issues and ensure returned value are correct
        self.assertEqual(assignee_count_response, assignee_count_answer)

    @mock.patch('jav.core.javJira')
    @mock.patch('jav.core.javConfig')
    def test_refresh_dailydata_cache(self, mock_config, mock_jira):
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')
        mock_jira.get_completed_tickets = mock.MagicMock(return_value=self.get_jira_issues())

        start_date = datetime.strptime('2017-05-28', '%Y-%m-%d').date()
        end_date = datetime.strptime('2017-05-21', '%Y-%m-%d').date()

        # App init, necessary to get to the logging service
        app = self.get_app()

        import_data = ImportData(app.log, mock_config)
        import_data.jira = mock_jira

        # To simplify formatting, re-run the full expected answer through the function
        refresh_daily_data_cache_response = import_data.refresh_dailydata_cache(self.get_data_completion(), start_date, end_date)

        self.assertDictEqual(refresh_daily_data_cache_response, self.get_data_completion_answer())

    @mock.patch('jav.core.javFiles.Files')
    @mock.patch('jav.core.javJira')
    @mock.patch('jav.core.javConfig')
    def test_get_remaining_work(self, mock_config, mock_jira, mock_files):
        mock_config.get_config_value = mock.MagicMock(return_value='jira_points_field')
        mock_jira.get_remaining_tickets = mock.MagicMock(return_value=self.get_jira_issues())
        mock_files.json_write = mock.MagicMock(return_value=True)

        # App init, necessary to get to the logging service
        app = self.get_app()

        import_data = ImportData(app.log, mock_config)
        import_data.jira = mock_jira
        import_data.files = mock_files

        get_remaining_work_answer = {
            'points': 37
            , 'tickets': 5
            , 'types': {
                'defect': {'type': 'defect', 'points': 30, 'tickets': 3}
                , 'story': {'type': 'story', 'points': 7, 'tickets': 2}
            }
            , 'assignees': {
                'johnd': {'displayName': 'John Doe', 'points': 10, 'tickets': 1}
                , 'brucew': {'displayName': 'Bruce Wayne', 'points': 22, 'tickets': 2}
                , 'darthv': {'displayName': 'Darth Vader', 'points': 5, 'tickets': 2}
            }
        }
        self.assertDictEqual(import_data.get_remaining_work(), get_remaining_work_answer)





from jav.core.javConfig import Config
from jav.core.javLogConfig import LogConfig


class Setup(object):
    """
        Used to setup the script (fill configuration settings)
    """

    def __init__(self, log, app_config):
        self.log = log
        self.config = Config(self.log)
        self.log_config = LogConfig(self.log, app_config, self.config.config_path + 'setup.log')

    def main(self):
        self.log.info('Initiating App Setup')

        # If conf init was called when Config class was initialized, then we don't call it again
        # This would happen if setup is called before a config file was created
        if self.config.config_init is False:
            self.config.init_config()



import collections
import dateutil.parser
import json
import os

class Files(object):
    """
        The main purpose of this function is to write/read from JSON and JSONL files
    """
    def __init__(self, log):
        self.log = log

    def json_load(self, filepath):
        self.log.info('Files.json_load(): Loading from file: ' + filepath)
        if os.path.isfile(filepath):
            with open(filepath) as file_to_open:
                data = json.load(file_to_open)
        else:
            data = None
        return data

    def json_write(self, filepath, content):
        self.log.info('Files.json_write(): Writing to file: ' + filepath)
        with open(filepath, "w") as json_file:
            json_file.write(json.dumps(content))
        return True

    def jsonl_load(self, filepath):
        self.log.info('Files.jsonl_load(): Loading from file: ' + filepath)
        # We consider here that any jsonl line does contain a datatime value
        data = collections.OrderedDict()
        if os.path.isfile(filepath):
            for line in open(filepath).readlines():
                current_line = json.loads(line)
                current_line['datetime'] = dateutil.parser.parse(current_line['datetime'])
                dict_idx = current_line['datetime'].strftime('%Y%m%d')
                data[dict_idx] = current_line

        return data

    def jsonl_write(self, filepath, content):
        self.log.info('Files.json_write(): Writing to file: ' + filepath)
        self.log.info(content)

    def jsonl_append(self, filepath, content):
        self.log.debug('Files.jsonl_append(): Append to file: ' + filepath)
        with open(filepath, 'a+') as file_to_write:
            file_to_write.write(json.dumps(content) + '\n')

    @staticmethod
    def prep_path(path):
        if path[-1] != '/':
            path = path + '/'
        if not os.path.isdir(path):
            os.makedirs(path)
        return path


"""CLI tests for jav."""

# from jav.utils import test
#
# class CliTestCase(test.javTestCase):
#     def test_jav_cli(self):
#         argv = ['--foo=bar']
#         with self.make_app(argv=argv) as app:
#             app.run()
#             self.eq(app.pargs.foo, 'bar')



import slackweb

class Msg(object):
    """
        Display messages to console or send to slack depending of the user selected mode
    """

    def __init__(self, log, config, send):
        self.log = log
        self.config = config
        self.__send = send
        if self.send:
            self.__slack = slackweb.Slack(url=self.config.get_config_value('slack_webhook'))

    @property
    def send(self):
        return self.__send

    @property
    def slack(self):
        return self.__slack

    def publish(self, stats_days, stats_weeks, stats_remaining):
        remaining = []
        for scan_day in stats_remaining:
            remaining = stats_remaining[scan_day]['days_to_completion']
            remaining_pts = stats_remaining[scan_day][self.config.get_config_value('stats_metric')]
            break
        for scan_day in stats_days:
            daily_velocity = stats_days[scan_day]
            daily_metric = stats_days[scan_day][self.config.get_config_value('stats_metric')]
            day_txt = stats_days[scan_day]['daytxt']
            break
        for scan_week in stats_weeks:
            week_txt = stats_weeks[scan_week]['weektxt']
            weekly_velocity = stats_weeks[scan_week]['stats']
            weekly_metric = stats_weeks[scan_week][self.config.get_config_value('stats_metric')]
            week_estimate = None
            if stats_weeks[scan_week]['days'] < 5:
                days_remaining = 5 - stats_weeks[scan_week]['days']
                points_remaining = days_remaining * stats_weeks[scan_week]['stats']['4']['avg']
                week_estimate = points_remaining + stats_weeks[scan_week]['points']
            break

        if daily_metric > daily_velocity['sameday']['4']['avg']:
            trend_day = ':arrow_upper_right:'
        elif daily_metric < daily_velocity['sameday']['4']['avg']:
            trend_day = ':arrow_lower_right:'
        else:
            trend_day = ':arrow_right:'

        if week_estimate is None:
            if weekly_metric > weekly_velocity['4']['avg']:
                trend_week = ':arrow_upper_right:'
            elif weekly_metric < weekly_velocity['4']['avg']:
                trend_week = ':arrow_lower_right:'
            else:
                trend_week = ':arrow_right:'
        else:
            if week_estimate > weekly_velocity['4']['avg']:
                trend_week = ':arrow_upper_right: (Est: ' + week_estimate + ')'
            elif week_estimate < weekly_velocity['4']['avg']:
                trend_week = ':arrow_lower_right: (Est: ' + week_estimate + ')'
            else:
                trend_week = ':arrow_right: (Est: ' + week_estimate + ')'

        if self.config.get_config_value('stats_metric') == 'tickets':
            metric_legend = 'Tickets'
            metric_short = 'tix'
        else:
            metric_legend = 'Story Points'
            metric_short = 'pts'

        self.slack_msg('Howdy everyone, here are our velocity stats, <'
                       + self.config.get_config_value('git_pageurl')
                       + '|live from Jira>.'
                       + '\n'
                       + 'Remaining ' + metric_legend + ': *'
                       + str(remaining_pts)
                       + '*\n'
                       + 'Completed on ' + day_txt + ': ' + str(daily_metric) + ' ' + metric_short + ' ['
                       + 'Max: ' + str(daily_velocity['sameday']['4']['max'])
                       + ' / '
                       + 'Min: ' + str(daily_velocity['sameday']['4']['min'])
                       + ' / '
                       + 'Avg: ' + str(daily_velocity['sameday']['4']['avg'])
                       + '] '
                       + trend_day
                       + '\n'
                       + 'Completed this week (' + week_txt + '): ' + str(weekly_metric) + ' ' + metric_short + ' ('
                       + 'Max: ' + str(weekly_velocity['4']['max'])
                       + ' / '
                       + 'Min: ' + str(weekly_velocity['4']['min'])
                       + ' / '
                       + 'Avg: ' + str(weekly_velocity['4']['avg'])
                       + ') '
                       + trend_week
                       + '\n'
                       + 'Days to Completion: *'
                       + str(round(remaining['4'], 1))
                       + ' days'
                       + '*\n'
                       + '_Most numbers are calculated over previous 4 weeks, excluding current day/week_'
                       )

    def slack_msg(self, msg):
        self.log.info(msg)
        if self.send:
            self.slack.notify(text=msg, channel=self.config.get_config_value('slack_channel'))



import subprocess
import json

class Github(object):
    """
    This class publish the chart to github pages
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config
        self.__repo = None

    @property
    def repo(self):
        return self.__repo

    @repo.setter
    def repo(self, value):
        self.__repo = value

    def sys_call(self, command):
        try:
            self.log.info('Github.sys_call(): Calling: ' + json.dumps(command))
            output = subprocess.check_output(command, stderr=subprocess.STDOUT)
            self.log.info(output.decode('utf-8'))
            return output
        except subprocess.CalledProcessError as e:
            self.log.error('Github.sys_call(): ' + e.output.decode('utf-8'))
            return False

    def git_clone(self):
        git_command = [
            'git'
            , 'clone'
            , self.config.get_config_value('git_repo')
            , self.config.get_config_value('git_localpath')
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_init(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'init'
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_commit(self, message):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , '--work-tree'
            , self.config.get_config_value('git_localpath')
            , 'commit'
            , '-m'
            , message
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_remote_add_origin(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'remote'
            , 'add'
            , 'origin'
            , self.config.git_repo
        ]
        if self.sys_call(git_command) is False:
            exit()


    def git_push_branch(self, branch):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'push'
            , '-u'
            , 'origin'
            , branch
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_push(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'push'
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_branch_upstream(self, branch):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , '--work-tree'
            , self.config.get_config_value('git_localpath')
            , 'branch'
            , '--set-upstream-to=origin/' + branch + ' ' + branch
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_pull_branch(self, branch, origin=False):
        if origin is True:
            branch = 'origin ' + branch
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , '--work-tree'
            , self.config.get_config_value('git_localpath')
            , 'pull'
            , branch
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_pull(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , '--work-tree'
            , self.config.get_config_value('git_localpath')
            , 'pull'
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_stash(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'stash'
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_add(self, filename):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , '--work-tree'
            , self.config.get_config_value('git_localpath')
            , 'add'
            , filename
        ]
        if self.sys_call(git_command) is False:
            exit()

    def git_checkout_create(self, branch_name):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'checkout'
            , '-b'
            , branch_name
        ]
        if self.sys_call(git_command) is False:
            exit()


    def git_checkout(self, branch_name):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'checkout'
            , branch_name
        ]
        return self.sys_call(git_command)

    def git_checkout_f(self, branch_name):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'checkout'
            , '-f'
            , branch_name
        ]
        return self.sys_call(git_command)

    def git_rm_cached(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'rm'
            , '--cached'
        ]
        if self.sys_call(git_command) is False:
            exit()


    def get_branches(self):
        git_command = [
            'git'
            , '--git-dir'
            , self.config.get_config_value('git_localpath') + '.git'
            , 'branch'
            , '-a'
        ]
        output = self.sys_call(git_command)
        if output is False:
            exit()

        branches = []
        for line in output.decode('utf-8').split('\n'):
            if len(line) > 3:
                self.log.info('Github.get_branches() - Branch: ' + line + ' found')
                branches.append(line)
        return branches


    def is_folder_git_repo(self):
        git_command = [
            'git'
            , '--git-dir=' + self.config.get_config_value('git_localpath') + '.git'
            , 'status'
        ]
        return self.sys_call(git_command)



from unittest import TestCase
from jav.core.javClear import Clear
import mock
from cement.core import foundation

class TestClear(TestCase):

    @classmethod
    def get_app(self):
        """App init, necessary to get to the logging service"""
        app = foundation.CementApp('myapp')
        app.setup()
        app.run()
        return app

    @mock.patch('jav.core.javConfig')
    def test_main(self, mock_config):
        """This is a dummy test, only verifies that the function exists and return True"""
        # App init, necessary to get to the logging service
        app = self.get_app()

        clear = Clear(app.log, mock_config)
        mock_config.filepath_data_completion = '/thispathdoesnotexist/'

        self.assertEqual(clear.main(), True)

    # @mock.patch('os.path.isdir')
    # @mock.patch('os.path.isfile')
    # @mock.patch('jav.core.javConfig.Config.load_config')
    # def test_get_config_value(self, mock_conf_load, mock_isfile, mock_isdir):
    #     """Receives a parameter and test if a value can be returned"""
    #     # App init, necessary to get to the logging service
    #     app = self.get_app()
    #     #mock_isfile.return_value = True
    #     #mock_isdir.return_value = True
    #     mock_conf_load.return_value = self.get_config()
    #
    #     config = Config(app.log, config_values = self.get_config())
    #
    #     self.assertEqual(config.get_config_value('jira_jql_remaining'), 'JQL QUERY')






from datetime import datetime
import pytz


class Time(object):
    """
        Class in charge or handling various time-related functions
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config

    @staticmethod
    def get_current_date():
        return datetime.now(pytz.timezone('America/Toronto'))

    def get_end_date(self):
        """Get end_date as a datetime object from configuration"""
        return datetime.strptime(self.config.get_config_value('end_date'), "%Y-%m-%d").date()




from jav.core.javFiles import Files
from jav.core.javStatsDay import StatsDay
from jav.core.javStatsRemaining import StatsRemaining
from jav.core.javStatsWeek import StatsWeek
from jav.core.javTime import Time


class Crunch(object):
    """
        Receive data and crunch numbers
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config
        self.time = Time(self.log, self.config)

    def crunch_stats(self, daily_data, remaining_work):
        self.log.info('Crunching statistics')

        # Calculate stats based on Jira Data
        stats_weeks = StatsWeek(self.log, self.config, daily_data).main()
        stats_days = StatsDay(self.log, self.config, daily_data).main()
        stats_remaining = StatsRemaining(self.log, self.config, self.time, stats_weeks, remaining_work).main()

        return stats_days, stats_weeks, stats_remaining

    def load_stats_cache(self):
        self.log.info('Loading previously calculated stats from cache')
        stats_days = Files(self.log).jsonl_load(self.config.filepath_stats_days)
        stats_weeks = Files(self.log).jsonl_load(self.config.filepath_stats_weeks)
        stats_remaining = Files(self.log).jsonl_load(self.config.filepath_stats_remaining)

        return stats_days, stats_weeks, stats_remaining



import collections
import numpy
import os
import copy
from jav.core.javFiles import Files


class StatsWeek(object):
    """
        The main function of this class goes through all the data points and break it down by week
        It then calculate various metrics for this particular week
        It also look in history to extract trend and evolution over time (such as evolution of the weekly story points
        or tickets average over time)
    """

    def __init__(self, log, config, daily_data):
        self.log = log
        self.config = config
        self.__daily_data = daily_data
        self.__weeks_data = collections.OrderedDict()

    @property
    def daily_data(self):
        return self.__daily_data

    @property
    def weeks_data(self):
        return self.__weeks_data

    def main(self):
        self.log.info('Calculate weekly stats throughout the captured period')

        # First pass, calculate weekly values
        for current_day in self.daily_data:
            week_txt = self.daily_data[current_day]['datetime'].strftime('%Y.%W')
            if week_txt not in self.weeks_data:
                self.weeks_data[week_txt] = {
                    'values': []
                    , 'datetime': self.daily_data[current_day]['datetime']
                    , 'weektxt': week_txt
                    , 'stats': {}
                }
            self.weeks_data[week_txt]['values'].append(
                self.daily_data[current_day][self.config.get_config_value('stats_metric')])
            self.weeks_data[week_txt]['days'] = len(self.weeks_data[week_txt]['values'])
            self.weeks_data[week_txt][self.config.get_config_value('stats_metric')] = sum(
                self.weeks_data[week_txt]['values'])

        # Second pass, get min and max since 'beginning of time'
        for week_txt in self.weeks_data:
            del self.weeks_data[week_txt]['values']
            week_found = False
            for scan_week in self.weeks_data:
                if week_found is True:
                    if 'all' not in self.weeks_data[week_txt]['stats']:
                        self.weeks_data[week_txt]['stats']['all'] = {
                            'values': []
                        }
                    self.weeks_data[week_txt]['stats']['all']['values'].append(
                        self.weeks_data[scan_week][self.config.get_config_value('stats_metric')])
                    self.weeks_data[week_txt]['stats']['all']['avg'] = int(
                        numpy.mean(self.weeks_data[week_txt]['stats']['all']['values']))
                    self.weeks_data[week_txt]['stats']['all']['max'] = max(
                        self.weeks_data[week_txt]['stats']['all']['values'])
                    self.weeks_data[week_txt]['stats']['all']['min'] = min(
                        self.weeks_data[week_txt]['stats']['all']['values'])
                    for week_idx in self.config.get_config_value('rolling_stats'):
                        if week_idx == 'all' or len(self.weeks_data[week_txt]['stats']['all']['values']) <= week_idx:
                            if week_idx not in self.weeks_data[week_txt]['stats']:
                                self.weeks_data[week_txt]['stats'][week_idx] = {
                                    'values': []
                                }
                            self.weeks_data[week_txt]['stats'][week_idx]['values'].append(
                                self.weeks_data[scan_week][self.config.get_config_value('stats_metric')])
                            self.weeks_data[week_txt]['stats'][week_idx]['avg'] = int(
                                numpy.mean(self.weeks_data[week_txt]['stats'][week_idx]['values']))
                            self.weeks_data[week_txt]['stats'][week_idx]['max'] = max(
                                self.weeks_data[week_txt]['stats'][week_idx]['values'])
                            self.weeks_data[week_txt]['stats'][week_idx]['min'] = min(
                                self.weeks_data[week_txt]['stats'][week_idx]['values'])

                if scan_week == week_txt:
                    week_found = True

        # Then write content to a JSONL file
        if os.path.isfile(self.config.filepath_stats_weeks):
            os.remove(self.config.filepath_stats_weeks)

        for week_txt in self.weeks_data:
            if 'all' in self.weeks_data[week_txt]['stats']:
                del self.weeks_data[week_txt]['stats']['all']['values']
            for week_idx in self.config.get_config_value('rolling_stats'):
                if week_idx in self.weeks_data[week_txt]['stats']:
                    del self.weeks_data[week_txt]['stats'][week_idx]['values']
            week_obj = copy.deepcopy(self.weeks_data[week_txt])
            week_obj['datetime'] = self.weeks_data[week_txt]['datetime'].isoformat()

            Files(self.log).jsonl_append(self.config.filepath_stats_weeks, week_obj)

        return self.weeks_data



"""Jira Agile Velocity base controller."""

from cement.ext.ext_argparse import ArgparseController, expose

from jav.core.javSetup import Setup
from jav.core.javClear import Clear
from jav.core.javCrunch import Crunch
from jav.core.javLoad import Load
from jav.core.javConfig import Config
from jav.core.javLogConfig import LogConfig
from jav.core.javBuildChart import BuildChart
from jav.core.javPublishGithubPage import PublishGithubPage
from jav.core.javMsg import Msg


class javBaseController(ArgparseController):

    class Meta:
        label = 'base'
        description = 'Connect to Jira REST API to collect completed story points, calculate weekly velocity, and estimate completion date'
        arguments = [
            (
                ['-s', '--send'],
                dict(help='Send message to slack, by default prints to console', action='store_true')
            ),
            (
                ['-p', '--path'],
                dict(help='Path to the directory storing data and config, will be created if does not exist', action='store', dest='path_config')
            )
        ]

    @expose(help='Clear previous data (USE WITH CAUTION)')
    def clear(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'clear.log')

        clear = Clear(self.app.log, config)
        clear.main()

    @expose(help='Enter setup mode and provide configuration parameters (jira creds, slack details)')
    def setup(self):
        setup = Setup(self.app.log, self.app.config)
        setup.main()

    @expose(help='Load latest data from Jira into cache')
    def load(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'load.log')

        load = Load(self.app.log, self.app.config)
        load.refresh_jira_cache()

    @expose(help='Obtain updated data from Jira and crunch numbers')
    def crunch(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'crunch.log')

        # Loading saved files into memory to be used by the component crunching numbers
        load = Load(self.app.log, self.app.config)
        daily_data, remaining_work = load.load_jira_cache()

        crunch = Crunch(self.app.log, config)
        crunch.crunch_stats(daily_data, remaining_work)

    @expose(help='Create charts from cached data')
    def chart(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'chart.log')

        #Get previously crunched number from cache file
        crunch = Crunch(self.app.log, config)
        stats_days, stats_weeks, stats_remaining = crunch.load_stats_cache()

        #Build charts
        BuildChart(self.app.log, config).main(stats_days, stats_weeks, stats_remaining)

    @expose(help='Publish charts to github pages')
    def publish(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'publish.log')

        PublishGithubPage(self.app.log, config).main()

    @expose(help='Send latest stats to the team on Slack')
    def msg(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'msg.log')

        #Get previously crunched number from cache file
        crunch = Crunch(self.app.log, config)
        stats_days, stats_weeks, stats_remaining = crunch.load_stats_cache()

        Msg(self.app.log, config, self.app.pargs.send).publish(stats_days, stats_weeks, stats_remaining)

    @expose(help='Get data, crunch numbers, do stuff')
    def run(self):
        config = Config(self.app.log, self.app.pargs.path_config)
        LogConfig(self.app.log, self.app.config, config.config_path + 'run.log')

        # Load data from Jira
        load = Load(self.app.log, self.app.config)
        daily_data, remaining_work = load.refresh_jira_cache()

        # Crunch numbers
        crunch = Crunch(self.app.log, config)
        stats_days, stats_weeks, stats_remaining = crunch.crunch_stats(daily_data, remaining_work)

        # Build Chart
        BuildChart(self.app.log, config).main(stats_days, stats_weeks, stats_remaining)

        # Publish Chart
        PublishGithubPage(self.app.log, config).main()

        #Get previously crunched number from cache file, to avoid the issue with json index key conversion
        # Issue, there is no numerical indexes in json, only strings.
        stats_days, stats_weeks, stats_remaining = crunch.load_stats_cache()

        # Message Team
        Msg(self.app.log, config, self.app.pargs.send).publish(stats_days, stats_weeks, stats_remaining)







from setuptools import setup, find_packages

setup(name='jav',
    version='0.1',
    description="Connect to Jira REST API to collect completed story points, calculate weekly velocity, and estimate completion date",
    long_description="Connect to Jira REST API to collect completed story points, calculate weekly velocity, and estimate completion date",
    classifiers=[],
    keywords='',
    author='fgerthoffert',
    author_email='contact@mail.gerthoffert.net',
    url='https://github.com/Fgerthoffert/jira-agile-velocity',
    license='GNU General Public License v3.0',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    test_suite='nose.collector',
    install_requires=[
        ### Required to build documentation
        # "Sphinx >= 1.0",
        ### Required for testing
        "nose",
        # "coverage",
        ### Required to function
        'cement',
        ],
    setup_requires=[],
    entry_points="""
        [console_scripts]
        jav = jav.cli.main:main
    """,
    namespace_packages=[],
    )



import collections
import os
from datetime import timedelta
from jav.core.javJira import Jira
import dateutil.parser
import copy
from jav.core.javFiles import Files

class ImportData(object):
    """ This class is used to obtain data used for processing

    Args:
        log: A class, the logging interface
        config: A class, the app config interface

    Attributes:
        tbc
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config
        self.__jira = Jira(self.log, self.config)
        self.__files = Files(self.log)

    @property
    def jira(self):
        return self.__jira

    @jira.setter
    def jira(self, jira_class):
        """This is mostly used for unittest to inject a mock Jira Class"""
        self.__jira = jira_class

    @property
    def files(self):
        return self.__files

    @files.setter
    def files(self, files_class):
        """This is mostly used for unittest to inject a mock Jira Class"""
        self.__files = files_class

    def write_dailydata_cache(self, daily_data):
        """Write an ordered dict into a JSONL file, converting datetime to isoformat"""
        self.log.info('ImportData.write_dailydata_cache(): Write back daily data cache to file: ' + self.config.filepath_data_completion)

        if os.path.isfile(self.config.filepath_data_completion):
            os.remove(self.config.filepath_data_completion)

        for currentdata in daily_data:
            daily_obj = copy.deepcopy(daily_data[currentdata])
            daily_obj['datetime'] = daily_data[currentdata]['datetime'].isoformat()
            self.files.jsonl_append(self.config.filepath_data_completion, daily_obj)

    def refresh_dailydata_cache(self, previous_data_cache, date_start, date_end):
        """
        Starting from date_start, loop through each day until date_end
        For each day, check if the day is already in previous_data cache:
         - If yes, skip the day
         - If not, get from Jira, completed tickets for that day
        """
        self.log.info('ImportData.refresh_dailydata_cache(): start')
        daily_data = collections.OrderedDict()

        date_current = date_start
        while 1:
            date_current = date_current - timedelta(days=1)
            dict_idx = date_current.strftime('%Y%m%d')

            item_found = False
            # We check if the day is already in the file
            for current_day_data in previous_data_cache:
                if date_current.strftime('%Y-%m-%d') == previous_data_cache[current_day_data]['datetime'].strftime(
                        '%Y-%m-%d'):
                    daily_obj = copy.deepcopy(previous_data_cache[current_day_data])
                    daily_obj['datetime'] = previous_data_cache[current_day_data]['datetime'].isoformat()
                    daily_data[dict_idx] = previous_data_cache[current_day_data]
                    self.log.info('ImportData.refresh_dailydata_cache(): ' + date_current.strftime(
                        '%Y.W%W-%A') + ': ' + date_current.strftime('%Y-%m-%d') + ' Already in cache')
                    item_found = True

            if not item_found:
                if date_current.strftime('%A') != 'Sunday' and date_current.strftime('%A') != 'Saturday':
                    self.log.info('ImportData.refresh_dailydata_cache(): ' + date_current.strftime(
                        '%Y.W%W-%A') + ': ' + date_current.strftime('%Y-%m-%d') + ' Obtaining daily data')
                    issues_list = self.jira.get_completed_tickets(date_current)
                    self.log.info('ImportData.refresh_dailydata_cache(): ' + date_current.strftime(
                        '%Y.W%W-%A') + ': ' + date_current.strftime('%Y-%m-%d') + ' Calculating stats')
                    daily_obj = {
                        'points': self.count_story_points(issues_list['issues'])
                        , 'tickets': len(issues_list['issues'])
                        , 'datetime': date_current.isoformat()
                        , 'types': self.story_types_count(issues_list['issues'])
                        , 'assignees': self.assignee_count(issues_list['issues'])
                    }
                    daily_data[dict_idx] = daily_obj
                    daily_data[dict_idx]['datetime'] = dateutil.parser.parse(daily_data[dict_idx]['datetime']).date()

            if date_current.strftime('%Y-%m-%d') < date_end.strftime('%Y-%m-%d'):
                self.log.info('ImportData.refresh_dailydata_cache(): All data collected')
                break
        return daily_data

    def get_story_points(self, issue):
        """Return the number of story points of a particular JIRA issue"""
        jira_points_field = self.config.get_config_value('jira_field_points')
        story_points = None
        try:
            story_points = int(issue['fields'][jira_points_field])
        except Exception:
            self.log.debug('Ticket missing story points')
            self.log.debug(issue)
        return story_points

    def story_types_count(self, issues_list):
        """Return a break down of story points and count per ticket type"""
        self.log.debug('ImportData.story_types_count(): Counting story points per ticket type')
        issues_types = {}
        for issue in issues_list:
            type_name = issue['fields']['issuetype']['name']
            if type_name not in issues_types:
                issues_types[type_name] = {'type': type_name, 'points': 0, 'tickets': 0}
            issue_points = self.get_story_points(issue)
            if issue_points is not None:
                issues_types[type_name]['points'] = int(issues_types[type_name]['points'] + issue_points)
            issues_types[type_name]['tickets'] = int(issues_types[type_name]['tickets'] + 1)
        return issues_types

    def assignee_count(self, issues_list):
        """Return a breakdown of story points and count per assignee """
        self.log.debug('ImportData.assignee_count(): Counting story points per assignees')
        assignees = {}
        for issue in issues_list:
            if issue['fields']['assignee'] is None:
                name = 'unassigned'
                name_display = 'Unassigned'
            else:
                name = issue['fields']['assignee']['name']
                name_display = issue['fields']['assignee']['displayName']
            if name not in assignees:
                assignees[name] = {'displayName': name_display, 'points': 0, 'tickets': 0}
            issue_points = self.get_story_points(issue)
            if issue_points is not None:
                assignees[name]['points'] = int(assignees[name]['points'] + issue_points)
            assignees[name]['tickets'] = int(assignees[name]['tickets'] + 1)
        return assignees

    def count_story_points(self, issues_list):
        """Count the total number of story points in a list of tickets"""
        self.log.debug('ImportData.count_story_points(): Counting total story points in a list of tickets')
        points = 0
        for issue in issues_list:
            issue_points = self.get_story_points(issue)
            if issue_points is not None:
                points = points + issue_points
        return int(points)

    def get_remaining_work(self):
        """Get remaining work from Jira and write the result to file"""
        self.log.info('ImportData.get_remaining_work(): Obtaining remaining work')
        issues_list = self.jira.get_remaining_tickets()

        remaining = {
            'points': self.count_story_points(issues_list['issues'])
            , 'tickets': len(issues_list['issues'])
            , 'types': self.story_types_count(issues_list['issues'])
            , 'assignees': self.assignee_count(issues_list['issues'])
        }
        self.files.json_write(self.config.filepath_data_remaining, remaining)
        return remaining



from unittest import TestCase
from jav.core.javConfig import Config
from cement.core import foundation

class TestConfig(TestCase):

    @classmethod
    def get_app(self):
        """App init, necessary to get to the logging service"""
        app = foundation.CementApp('myapp')
        app.setup()
        app.run()
        return app

    @classmethod
    def get_config(self):
        config = {
            'end_date': '2017-02-01'
            , 'git_branch': 'gh-pages'
            , 'git_localpath': '/Users/USERNAME/Desktop/test-agile-page/'
            , 'git_pageurl': 'https://org.github.io/repo/dir/'
            , 'git_pathdirectory': 'dir/'
            , 'git_repo': 'https://github.com/org/repo.git'
            , 'jira_field_points': 'customfield_10002'
            , 'jira_host': 'https://URL'
            , 'jira_jql_remaining': 'JQL QUERY'
            , 'jira_jql_velocity': 'JQL QUERY'
            , 'jira_password': 'PASSWORD'
            , 'jira_username': 'USERNAME'
            , 'rolling_stats': [4, 8, 12]
            , 'slack_channel': '#general'
            , 'slack_webhook': 'SLACK WEBHOOK'
            , 'stats_metric': 'tickets'
        }
        return config

    def test_get_config_value(self):
        """Receives a parameter and test if a value can be returned"""
        # App init, necessary to get to the logging service
        app = self.get_app()
        config = Config(app.log, config_values = self.get_config())

        self.assertEqual(config.get_config_value('jira_jql_remaining'), 'JQL QUERY')

    def test_set_config_value(self):
        """Modifies configuration of a parameter and check output"""
        # App init, necessary to get to the logging service
        app = self.get_app()
        config = Config(app.log, config_values = self.get_config())

        self.assertEqual(config.get_config_value('jira_jql_remaining'), 'JQL QUERY')
        self.assertEqual(config.set_config_value('jira_jql_remaining', 'NEW JQL QUERY'), 'NEW JQL QUERY')
        self.assertEqual(config.get_config_value('jira_jql_remaining'), 'NEW JQL QUERY')

    def test_init_config_auto(self):
        """Automatically initialize the default config"""
        # App init, necessary to get to the logging service
        app = self.get_app()
        config = Config(app.log, config_values=self.get_config())
        self.assertEqual(config.get_config_value('jira_jql_remaining'), 'JQL QUERY')
        config.init_config_auto()
        self.assertEqual(config.get_config_value('jira_jql_remaining'), 'JQL QUERY REMAINING')




from jav.core.javConfig import Config
from jav.core.javLogConfig import LogConfig
from jav.core.javImportData import ImportData
from jav.core.javTime import Time
from jav.core.javFiles import Files


class Load(object):
    """
        Load updated Jira data into cache
    """
    def __init__(self, log, app_config):
        self.log = log
        self.config = Config(self.log)
        self.time = Time(self.log, self.config)
        self.log_config = LogConfig(self.log, app_config, self.config.config_path + 'load.log')

    def refresh_jira_cache(self):
        self.log.info('Load updated data from Jira into Cache')

        date_start = self.time.get_current_date()
        date_end = self.time.get_end_date()
        self.log.info('Load.main(): Start Date: ' + date_start.strftime('%Y-%m-%d'))
        self.log.info('Load.main(): End Date: ' + date_end.strftime('%Y-%m-%d'))

        loader = ImportData(self.log, self.config)
        # Import existing data (if any) into a Python object
        previous_data = Files(self.log).jsonl_load(self.config.filepath_data_completion)
        # Refresh the cache by checking if additional days can be added
        daily_data = loader.refresh_dailydata_cache(previous_data, date_start, date_end)
        # Write back the data cache to file after clearing any existing one
        loader.write_dailydata_cache(daily_data)

        # Call Jira to get Remaining work
        remaining_work = loader.get_remaining_work()

        return daily_data, remaining_work

    def load_jira_cache(self):
        self.log.info('Load daily data and remaining work from cache')
        # Import existing data (if any) into a Python object
        daily_data = Files(self.log).jsonl_load(self.config.filepath_data_completion)
        remaining_work = Files(self.log).json_load(self.config.filepath_data_remaining)
        if remaining_work is None or daily_data is None:
            self.log.error('Unable to load cached data, please run \'jav load\' first')
            exit()

        return daily_data, remaining_work




import collections
import numpy
import os
import copy
from jav.core.javFiles import Files


class StatsDay(object):
    """
        The main function of this class goes through all the data points and break it down by day
        It then calculate various metrics for this particular week
        It also look in history to extract trend and evolution over time (such as evolution of the weekly story points
        or tickets average over time)
    """

    def __init__(self, log, config, daily_data):
        self.log = log
        self.config = config
        self.__daily_data = daily_data
        self.__days = collections.OrderedDict()

    @property
    def daily_data(self):
        return self.__daily_data

    @property
    def days(self):
        return self.__days

    def main(self):
        self.log.info('Calculate daily stats throughout the captured period')

        for current_day in self.daily_data:
            day_txt = self.daily_data[current_day]['datetime'].strftime('%Y%m%d')
            self.days[day_txt] = {
                'datetime': self.daily_data[current_day]['datetime']
                , 'daytxt': self.daily_data[current_day]['datetime'].strftime('%A')
                , 'points': self.daily_data[current_day]['points']
                , 'tickets': self.daily_data[current_day]['tickets']
                , 'anyday': {'all': {'values': []}}
                , 'sameday': {'all': {'values': []}}
            }
            for week_idx in self.config.get_config_value('rolling_stats'):
                if week_idx not in self.days[day_txt]['sameday']:
                    self.days[day_txt]['anyday'][week_idx] = {'values': []}
                    self.days[day_txt]['sameday'][week_idx] = {'values': []}

            day_found = False
            for scan_day in self.daily_data:
                if day_found:
                    for stats_type in ['anyday', 'sameday']:
                        same_weekday = False
                        if stats_type == 'anyday':
                            self.days[day_txt][stats_type]['all']['values'].append(
                                self.daily_data[scan_day][self.config.get_config_value('stats_metric')])
                        elif self.daily_data[current_day]['datetime'].strftime('%A') == self.daily_data[scan_day][
                            'datetime'].strftime('%A') and stats_type == 'sameday':
                            same_weekday = True
                            self.days[day_txt][stats_type]['all']['values'].append(
                                self.daily_data[scan_day][self.config.get_config_value('stats_metric')])

                        if self.days[day_txt][stats_type]['all']['values']:
                            self.days[day_txt][stats_type]['all']['avg'] = int(
                                numpy.mean(self.days[day_txt][stats_type]['all']['values']))
                            self.days[day_txt][stats_type]['all']['min'] = min(
                                self.days[day_txt][stats_type]['all']['values'])
                            self.days[day_txt][stats_type]['all']['max'] = max(
                                self.days[day_txt][stats_type]['all']['values'])

                        for week_idx in self.config.get_config_value('rolling_stats'):
                            in_range = False
                            if stats_type == 'anyday' and len(
                                    self.days[day_txt][stats_type]['all']['values']) <= week_idx * 5:
                                in_range = True
                            elif stats_type == 'sameday' and len(
                                    self.days[day_txt][stats_type]['all']['values']) <= week_idx:
                                in_range = True
                            if (stats_type == 'anyday' and in_range is True) or (
                                            in_range is True and same_weekday is True):
                                self.days[day_txt][stats_type][week_idx]['values'].append(
                                    self.daily_data[scan_day][self.config.get_config_value('stats_metric')])
                                self.days[day_txt][stats_type][week_idx]['avg'] = int(
                                    numpy.mean(self.days[day_txt][stats_type]['all']['values']))
                                self.days[day_txt][stats_type][week_idx]['min'] = min(
                                    self.days[day_txt][stats_type]['all']['values'])
                                self.days[day_txt][stats_type][week_idx]['max'] = max(
                                    self.days[day_txt][stats_type]['all']['values'])

                if self.daily_data[current_day]['datetime'].date() == self.daily_data[scan_day]['datetime'].date():
                    day_found = True

        # Then write content to a JSONL file
        if os.path.isfile(self.config.filepath_stats_days):
            os.remove(self.config.filepath_stats_days)

        # Clear un-necessary array values, and write output to a JSONL file
        for current_day in self.days:
            del self.days[current_day]['anyday']['all']['values']
            del self.days[current_day]['sameday']['all']['values']
            for week_idx in self.config.get_config_value('rolling_stats'):
                for stats_type in ['anyday', 'sameday']:
                    if stats_type in self.days[current_day] and week_idx in self.days[current_day][
                        stats_type] and 'values' in self.days[current_day][stats_type][week_idx]:
                        del self.days[current_day][stats_type][week_idx]['values']

            day_obj = copy.deepcopy(self.days[current_day])
            day_obj['datetime'] = self.days[current_day]['datetime'].isoformat()
            Files(self.log).jsonl_append(self.config.filepath_stats_days, day_obj)

        return self.days



import collections
import os
import copy
from jav.core.javFiles import Files

class StatsRemaining(object):
    """
        The main function of this class is to look at remaining work and guesstimate likely days to completion
    """

    def __init__(self, log, config, time, stats_weeks, remaining_work):
        self.log = log
        self.config = config
        self.time = time
        self.__stats_weeks = stats_weeks
        self.__remaining_work = remaining_work
        self.__remainingstats_filepath = self.config.config_path + 'stats_remaining.jsonl'

    @property
    def remaining_work(self):
        return self.__remaining_work

    def main(self):
        """Guesstimate efforts to completion, add result to a log file"""
        self.log.info('Guesstimating days to completion')
        for scan_week in self.__stats_weeks:
            current_week = self.__stats_weeks[scan_week]
            break

        date_current = self.time.get_current_date()

        remaining = {
            self.config.get_config_value('stats_metric'): self.remaining_work[self.config.get_config_value('stats_metric')]
            , 'datetime': date_current
            , 'types': self.remaining_work['types']
            , 'assignees': self.remaining_work['assignees']
            , 'days_to_completion': {}
        }
        self.log.info('Remaining number of story ' + self.config.get_config_value('stats_metric') + ': ' + str(remaining[self.config.get_config_value('stats_metric')]))
        for week_stat in current_week['stats']:
            avg_elements_per_day = float(current_week['stats'][week_stat]['avg']) / 5
            if avg_elements_per_day > 0:
                remaining_days = round(float(remaining[self.config.get_config_value('stats_metric')]) / float(avg_elements_per_day), 1)
            else:
                remaining_days = 0
            self.log.info('Over period: ' + str(week_stat) + ', average ' + self.config.get_config_value('stats_metric') + ' per day was: ' + str(
                avg_elements_per_day) + ' should completed in: ' + str(remaining_days) + ' days')
            remaining['days_to_completion'][week_stat] = remaining_days

        current_element_per_day = current_week[self.config.get_config_value('stats_metric')] / current_week['days']
        remaining_days = round(remaining[self.config.get_config_value('stats_metric')] / current_element_per_day, 1)
        remaining['days_to_completion']['current'] = remaining_days
        self.log.info('This week, average ' + self.config.get_config_value('stats_metric') + ' per day is: ' + str(
            current_element_per_day) + ' should completed in: ' + str(remaining_days) + ' days')

        # Load previous data and add current day (if already there, replace)
        daily_data = collections.OrderedDict()
        dict_idx = date_current.strftime('%Y-%m-%d')
        daily_data[dict_idx] = remaining

        current_data = Files(self.log).jsonl_load(self.config.filepath_stats_remaining)

        for current_day_data in current_data:
            if date_current.strftime('%Y-%m-%d') != current_data[current_day_data]['datetime'].strftime('%Y-%m-%d'):
                dict_idx = current_data[current_day_data]['datetime'].strftime('%Y-%m-%d')
                daily_data[dict_idx] = current_data[current_day_data]

        if os.path.isfile(self.config.filepath_stats_remaining):
            os.remove(self.config.filepath_stats_remaining)

        for currentdata in daily_data:
            daily_obj = copy.deepcopy(daily_data[currentdata])
            daily_obj['datetime'] = daily_data[currentdata]['datetime'].isoformat()
            self.log.info('StatsRemaining.main(): Writing stats for date: ' + daily_obj['datetime'])
            Files(self.log).jsonl_append(self.config.filepath_stats_remaining, daily_obj)

        return daily_data



from jav.core.javFiles import Files
from jav.core.javGithub import Github
from jav.core.javTime import Time
import os
import shutil


class PublishGithubPage(object):
    """
    This class publish the chart to github pages
    """

    def __init__(self, log, config):
        self.log = log
        self.config = config

        self.github = Github(self.log, self.config)
        self.time = Time(self.log, self.config)

    def main(self):
        """This hacky function publishes chart to github pages"""
        self.log.info('Publishing Website')
        self.log.info('Git local path: ' + self.config.get_config_value('git_localpath'))
        self.config.set_config_value('git_localpath', Files.prep_path(self.config.get_config_value('git_localpath')))

        if not self.github.is_folder_git_repo():
            self.log.info('Cloning Git Repository: ' + self.config.get_config_value('git_repo'))
            self.github.git_clone()
        else:
            self.log.info('Repository already available on local filesystem: ' + self.config.get_config_value('git_repo'))
            if self.github.git_pull() is not True:
                self.log.info('Git Stash')
                self.github.git_stash()

        # Test if requested branch exists, if not, create
        # Then switch to that branch
        remote_branches = self.github.get_branches()
        if not any(self.config.get_config_value('git_branch') in s for s in remote_branches):
            # Init branch
            self.log.warning(
                'The remote repository does not contain branch: ' + self.config.get_config_value('git_branch'))
            self.github.git_checkout_create(self.config.get_config_value('git_branch'))
            self.github.git_pull_branch(self.config.get_config_value('git_branch'), True)
            self.github.git_push_branch(self.config.get_config_value('git_branch'))
        elif any('* ' + self.config.get_config_value('git_branch') in s for s in remote_branches):
            self.log.info('Local repository already selected branch: ' + self.config.get_config_value('git_branch'))
        else:
            self.log.info('Switching to branch: ' + self.config.get_config_value('git_branch'))
            self.github.git_checkout_f(self.config.get_config_value('git_branch'))

        # Next do one last pull and copy content
        self.github.git_stash()

        # Then copy chart to requested directory
        if not os.path.isfile(self.config.filepath_charts + 'index.html'):
            self.log.error('Unable to locate source chart file at: ' + self.config.filepath_charts + 'index.html')
            exit()

        dst_path = Files.prep_path(
            self.config.get_config_value('git_localpath') + self.config.get_config_value('git_pathdirectory'))
        self.log.info('Preparing to copy chart file to: ' + dst_path)
        shutil.copyfile(self.config.filepath_charts + 'index.html', dst_path + 'index.html')
        self.log.info('Chart file copied')
        self.github.git_add('--all')
        self.github.git_commit('Copied chart file - ' + self.time.get_current_date().isoformat())
        self.github.git_push()
        self.log.info('Chart pushed to: ' + self.config.get_config_value('git_pageurl'))






"""Jira Agile Velocity main application entry point."""

from cement.core.foundation import CementApp
from cement.utils.misc import init_defaults
from cement.core.exc import FrameworkError, CaughtSignal
from jav.core import exc

# Application default.  Should update config/jav.conf to reflect any
# changes, or additions here.
defaults = init_defaults('jav')

# All internal/external plugin configurations are loaded from here
defaults['jav']['plugin_config_dir'] = '/etc/jav/plugins.d'

# External plugins (generally, do not ship with application code)
defaults['jav']['plugin_dir'] = '/var/lib/jav/plugins'

# External templates (generally, do not ship with application code)
defaults['jav']['template_dir'] = '/var/lib/jav/templates'


class javApp(CementApp):
    class Meta:
        label = 'jav'
        config_defaults = defaults

        # All built-in application bootstrapping (always run)
        bootstrap = 'jav.cli.bootstrap'

        # Internal plugins (ship with application code)
        plugin_bootstrap = 'jav.cli.plugins'

        # Internal templates (ship with application code)
        template_module = 'jav.cli.templates'

        # call sys.exit() when app.close() is called
        exit_on_close = True


class javTestApp(javApp):
    """A test app that is better suited for testing."""
    class Meta:
        # default argv to empty (don't use sys.argv)
        argv = []

        # don't look for config files (could break tests)
        config_files = []

        # don't call sys.exit() when app.close() is called in tests
        exit_on_close = False


# Define the applicaiton object outside of main, as some libraries might wish
# to import it as a global (rather than passing it into another class/func)
app = javApp()

def main():
    with app:
        try:
            app.run()

        except exc.javError as e:
            # Catch our application errors and exit 1 (error)
            print('javError > %s' % e)
            app.exit_code = 1

        except FrameworkError as e:
            # Catch framework errors and exit 1 (error)
            print('FrameworkError > %s' % e)
            app.exit_code = 1

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('CaughtSignal > %s' % e)
            app.exit_code = 0


if __name__ == '__main__':
    main()
